@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include "config.h"
@@ -69,14 +69,12 @@ poppler_page_finalize (GObject *object)
   g_object_unref (page->document);
   page->document = NULL;
 
+  if (page->annots != NULL)
+    delete page->annots;
   if (page->gfx != NULL)
     delete page->gfx;
   if (page->text_dev != NULL)
     delete page->text_dev;
-#if defined (HAVE_CAIRO)
-  if (page->image_dev != NULL)
-    delete page->image_dev;
-#endif
   /* page->page is owned by the document */
 }
 
@@ -232,8 +230,34 @@ poppler_page_get_transition (PopplerPage *page)
   return transition;
 }
 
+static TextOutputDev *
+poppler_page_get_text_output_dev (PopplerPage *page)
+{
+  if (page->text_dev == NULL) {
+    page->text_dev = new TextOutputDev (NULL, gTrue, gFalse, gFalse);
+
+    if (page->gfx)
+      delete page->gfx;
+    page->gfx = page->page->createGfx(page->text_dev,
+				      72.0, 72.0, 0,
+				      gFalse, /* useMediaBox */
+				      gTrue, /* Crop */
+				      -1, -1, -1, -1,
+				      gFalse, /* printing */
+				      page->document->doc->getCatalog (),
+				      NULL, NULL, NULL, NULL);
+
+    page->page->display(page->gfx);
+
+    page->text_dev->endPage();
+  }
+
+  return page->text_dev;
+}
+
 #if defined (HAVE_CAIRO)
 
+#ifdef POPPLER_WITH_GDK
 typedef struct {
   unsigned char *cairo_data;
   cairo_surface_t *surface;
@@ -340,7 +364,18 @@ poppler_page_copy_to_pixbuf (PopplerPage *page,
   gfree (output_dev_data->cairo_data);
 }
 
-#elif defined (HAVE_SPLASH)
+static void
+poppler_page_set_selection_alpha (PopplerPage           *page,
+				  double                 scale,
+				  GdkPixbuf             *pixbuf,
+				  PopplerSelectionStyle  style,
+				  PopplerRectangle      *selection)
+{
+  /* Cairo doesn't need this, since cairo generates an alpha channel. */ 
+}
+#endif /* POPPLER_WITH_GDK */
+
+#elif defined (HAVE_SPLASH) && defined (POPPLER_WITH_GDK)
  
 typedef struct {
 } OutputDevData;
@@ -404,182 +439,6 @@ poppler_page_copy_to_pixbuf(PopplerPage *page,
   delete [] pixel;
 }
 
-#endif
-
-#if defined (HAVE_CAIRO)
-
-/**
- * poppler_page_render:
- * @page: the page to render from
- * @cairo: cairo context to render to
- *
- * Render the page to the given cairo context.
- **/
-void
-poppler_page_render (PopplerPage *page,
-		     cairo_t *cairo)
-{
-  CairoOutputDev *output_dev;
-
-  g_return_if_fail (POPPLER_IS_PAGE (page));
-
-  output_dev = page->document->output_dev;
-  output_dev->setCairo (cairo);
-
-  /* NOTE: instead of passing -1 we should/could use cairo_clip_extents()
-   * to get a bounding box */
-  page->page->displaySlice(output_dev,
-			   72.0, 72.0, 0,
-			   gFalse, /* useMediaBox */
-			   gTrue, /* Crop */
-			   -1, -1,
-			   -1, -1,
-			   gFalse, /* printing */
-			   page->document->doc->getCatalog ());
-
-  output_dev->setCairo (NULL);
-}
-
-#endif
-
-/**
- * poppler_page_render_to_pixbuf:
- * @page: the page to render from
- * @src_x: x coordinate of upper left corner  
- * @src_y: y coordinate of upper left corner  
- * @src_width: width of rectangle to render  
- * @src_height: height of rectangle to render
- * @scale: scale specified as pixels per point
- * @rotation: rotate the document by the specified degree
- * @pixbuf: pixbuf to render into
- *
- * First scale the document to match the specified pixels per point,
- * then render the rectangle given by the upper left corner at
- * (src_x, src_y) and src_width and src_height.
- **/
-void
-poppler_page_render_to_pixbuf (PopplerPage *page,
-			       int src_x, int src_y,
-			       int src_width, int src_height,
-			       double scale,
-			       int rotation,
-			       GdkPixbuf *pixbuf)
-{
-  OutputDevData data;
-
-  g_return_if_fail (POPPLER_IS_PAGE (page));
-  g_return_if_fail (scale > 0.0);
-  g_return_if_fail (pixbuf != NULL);
-
-  poppler_page_prepare_output_dev (page, scale, rotation, FALSE, &data);
-
-  page->page->displaySlice(page->document->output_dev,
-			   72.0 * scale, 72.0 * scale,
-			   rotation,
-			   gFalse, /* useMediaBox */
-			   gTrue, /* Crop */
-			   src_x, src_y,
-			   src_width, src_height,
-			   gFalse, /* printing */
-			   page->document->doc->getCatalog ());
-  
-  poppler_page_copy_to_pixbuf (page, pixbuf, &data);
-}
-
-static TextOutputDev *
-poppler_page_get_text_output_dev (PopplerPage *page)
-{
-  if (page->text_dev == NULL) {
-    page->text_dev = new TextOutputDev (NULL, gTrue, gFalse, gFalse);
-
-    if (page->gfx)
-      delete page->gfx;
-    page->gfx = page->page->createGfx(page->text_dev,
-				      72.0, 72.0, 0,
-				      gFalse, /* useMediaBox */
-				      gTrue, /* Crop */
-				      -1, -1, -1, -1,
-				      gFalse, /* printing */
-				      page->document->doc->getCatalog (),
-				      NULL, NULL, NULL, NULL);
-
-    page->page->display(page->gfx);
-
-    page->text_dev->endPage();
-  }
-
-  return page->text_dev;
-}
-
-/**
- * poppler_page_get_selection_region:
- * @page: a #PopplerPage
- * @scale: scale specified as pixels per point
- * @style: a #PopplerSelectionStyle
- * @selection: start and end point of selection as a rectangle
- * 
- * Returns a region containing the area that would be rendered by
- * poppler_page_render_selection() or
- * poppler_page_render_selection_to_pixbuf(). The returned
- * region must be freed with gdk_region_destroy().
- * 
- * Return value: a newly allocated #GdkRegion
- **/
-GdkRegion *
-poppler_page_get_selection_region (PopplerPage           *page,
-				   gdouble                scale,
-				   PopplerSelectionStyle  style,
-				   PopplerRectangle      *selection)
-{
-  TextOutputDev *text_dev;
-  PDFRectangle poppler_selection;
-  SelectionStyle selection_style = selectionStyleGlyph;
-  GooList *list;
-  GdkRectangle rect;
-  GdkRegion *region;
-  int i;
-
-  poppler_selection.x1 = selection->x1;
-  poppler_selection.y1 = selection->y1;
-  poppler_selection.x2 = selection->x2;
-  poppler_selection.y2 = selection->y2;
-
-  switch (style)
-    {
-      case POPPLER_SELECTION_GLYPH:
-        selection_style = selectionStyleGlyph;
-	break;
-      case POPPLER_SELECTION_WORD:
-        selection_style = selectionStyleWord;
-	break;
-      case POPPLER_SELECTION_LINE:
-        selection_style = selectionStyleLine;
-	break;
-    }
-	      
-  text_dev = poppler_page_get_text_output_dev (page);
-  list = text_dev->getSelectionRegion(&poppler_selection, 
-				      selection_style, scale);
-
-  region = gdk_region_new();
-
-  for (i = 0; i < list->getLength(); i++) {
-    PDFRectangle *selection_rect = (PDFRectangle *) list->get(i);
-    rect.x      = (gint) selection_rect->x1;
-    rect.y      = (gint) selection_rect->y1;
-    rect.width  = (gint) (selection_rect->x2 - selection_rect->x1);
-    rect.height = (gint) (selection_rect->y2 - selection_rect->y1);
-    gdk_region_union_with_rect (region, &rect);
-    delete selection_rect;
-  }
-
-  delete list;
-
-  return region;
-}
-
-#if defined (HAVE_CAIRO)
-
 static void
 poppler_page_set_selection_alpha (PopplerPage           *page,
 				  double                 scale,
@@ -587,21 +446,8 @@ poppler_page_set_selection_alpha (PopplerPage           *page,
 				  PopplerSelectionStyle  style,
 				  PopplerRectangle      *selection)
 {
-  /* Cairo doesn't need this, since cairo generates an alpha channel. */ 
-}
-
-#elif defined (HAVE_SPLASH)
-
-static void
-poppler_page_set_selection_alpha (PopplerPage           *page,
-				  double                 scale,
-				  GdkPixbuf             *pixbuf,
-				  PopplerSelectionStyle  style,
-				  PopplerRectangle      *selection)
-{
-  GdkRegion *region;
-  gint n_rectangles, i, x, y, width, height;
-  GdkRectangle *rectangles;
+  GList *region, *l;
+  gint x, y, width, height;
   int pixbuf_rowstride, pixbuf_n_channels;
   guchar *pixbuf_data, *dst;
 
@@ -624,26 +470,176 @@ poppler_page_set_selection_alpha (PopplerPage           *page,
 
   region = poppler_page_get_selection_region (page, scale, style, selection);
 
-  gdk_region_get_rectangles (region, &rectangles, &n_rectangles);
-  for (i = 0; i < n_rectangles; i++) {
-    for (y = 0; y < rectangles[i].height; y++) {
-      dst = pixbuf_data + (rectangles[i].y + y) * pixbuf_rowstride +
-	rectangles[i].x * pixbuf_n_channels;
-      for (x = 0; x < rectangles[i].width; x++) {
+  for (l = region; l; l = g_list_next (l)) {
+    PopplerRectangle *rectangle = (PopplerRectangle *)l->data;
+    GdkRectangle rect;
+    
+    rect.x = (gint)rectangle->x1;
+    rect.y = (gint)rectangle->y1;
+    rect.width  = (gint) (rectangle->x2 - rectangle->x1);
+    rect.height = (gint) (rectangle->y2 - rectangle->y1);
+    
+    for (y = 0; y < rect.height; y++) {
+      dst = pixbuf_data + (rect.y + y) * pixbuf_rowstride +
+	rect.x * pixbuf_n_channels;
+      for (x = 0; x < rect.width; x++) {
 	  dst[3] = 0xff;
 	  dst += pixbuf_n_channels;
       }
     }
   }
 
-  g_free (rectangles);
-
-  gdk_region_destroy (region);
+  poppler_page_selection_region_free (region);
 }
 
-#endif
+#endif /* HAVE_SPLASH */
+
+static GBool
+poppler_print_annot_cb (Annot *annot, void *user_data)
+{
+  return (annot->getType() == Annot::typeWidget);
+}
 
 #if defined (HAVE_CAIRO)
+
+static void
+_poppler_page_render (PopplerPage *page,
+		      cairo_t *cairo,
+		      GBool printing)
+{
+  CairoOutputDev *output_dev;
+
+  g_return_if_fail (POPPLER_IS_PAGE (page));
+
+  output_dev = page->document->output_dev;
+  output_dev->setCairo (cairo);
+  output_dev->setPrinting (printing);
+
+  /* NOTE: instead of passing -1 we should/could use cairo_clip_extents()
+   * to get a bounding box */
+  page->page->displaySlice(output_dev,
+			   72.0, 72.0, 0,
+			   gFalse, /* useMediaBox */
+			   gTrue, /* Crop */
+			   -1, -1,
+			   -1, -1,
+			   printing,
+			   page->document->doc->getCatalog (),
+			   NULL, NULL,
+			   printing ? poppler_print_annot_cb : NULL, NULL);
+
+  output_dev->setCairo (NULL);	
+}
+
+/**
+ * poppler_page_render:
+ * @page: the page to render from
+ * @cairo: cairo context to render to
+ *
+ * Render the page to the given cairo context. This function
+ * is for rendering a page that will be displayed. If you want
+ * to render a page that will be printed use
+ * poppler_page_render_for_printing() instead
+ **/
+void
+poppler_page_render (PopplerPage *page,
+		     cairo_t *cairo)
+{
+  g_return_if_fail (POPPLER_IS_PAGE (page));
+
+  _poppler_page_render (page, cairo, gFalse);
+}
+
+/**
+ * poppler_page_render_for_printing:
+ * @page: the page to render from
+ * @cairo: cairo context to render to
+ *
+ * Render the page to the given cairo context for printing.
+ **/
+void
+poppler_page_render_for_printing (PopplerPage *page,
+				  cairo_t *cairo)
+{
+  g_return_if_fail (POPPLER_IS_PAGE (page));
+  
+  _poppler_page_render (page, cairo, gTrue);	
+}
+
+static cairo_surface_t *
+create_surface_from_thumbnail_data (guchar *data,
+				    gint    width,
+				    gint    height,
+				    gint    rowstride)
+{
+  guchar *cairo_pixels;
+  cairo_surface_t *surface;
+  static cairo_user_data_key_t key;
+  int j;
+
+  cairo_pixels = (guchar *)g_malloc (4 * width * height);
+  surface = cairo_image_surface_create_for_data ((unsigned char *)cairo_pixels,
+						 CAIRO_FORMAT_RGB24,
+						 width, height, 4 * width);
+  cairo_surface_set_user_data (surface, &key,
+			       cairo_pixels, (cairo_destroy_func_t)g_free);
+
+  for (j = height; j; j--) {
+    guchar *p = data;
+    guchar *q = cairo_pixels;
+    guchar *end = p + 3 * width;
+	  
+    while (p < end) {
+#if G_BYTE_ORDER == G_LITTLE_ENDIAN
+      q[0] = p[2];
+      q[1] = p[1];
+      q[2] = p[0];
+#else	  
+      q[1] = p[0];
+      q[2] = p[1];
+      q[3] = p[2];
+#endif
+      p += 3;
+      q += 4;
+    }
+
+    data += rowstride;
+    cairo_pixels += 4 * width;
+  }
+
+  return surface;
+}
+				    
+
+/**
+ * poppler_page_get_thumbnail:
+ * @page: the #PopperPage to get the thumbnail for
+ * 
+ * Get the embedded thumbnail for the specified page.  If the document
+ * doesn't have an embedded thumbnail for the page, this function
+ * returns %NULL.
+ * 
+ * Return value: the tumbnail as a cairo_surface_t or %NULL if the document
+ * doesn't have a thumbnail for this page.
+ **/
+cairo_surface_t *
+poppler_page_get_thumbnail (PopplerPage *page)
+{
+  unsigned char *data;
+  int width, height, rowstride;
+  cairo_surface_t *surface;
+
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
+
+  if (!page->page->loadThumb (&data, &width, &height, &rowstride))
+    return NULL;
+
+  surface = create_surface_from_thumbnail_data (data, width, height, rowstride);
+  gfree (data);
+  
+  return surface;
+}
+
 /**
  * poppler_page_render_selection:
  * @page: the #PopplerPage for which to render selection
@@ -669,8 +665,8 @@ poppler_page_render_selection (PopplerPage           *page,
 			       PopplerRectangle      *selection,
 			       PopplerRectangle      *old_selection,
 			       PopplerSelectionStyle  style, 
-			       GdkColor              *glyph_color,
-			       GdkColor              *background_color)
+			       PopplerColor          *glyph_color,
+			       PopplerColor          *background_color)
 {
   TextOutputDev *text_dev;
   CairoOutputDev *output_dev;
@@ -725,7 +721,136 @@ poppler_page_render_selection (PopplerPage           *page,
    * so we don't have to use TextOutputDev and render a second
    * time. */
 }
-#endif
+
+#endif /* HAVE_CAIRO */
+
+#ifdef POPPLER_WITH_GDK
+static void
+_poppler_page_render_to_pixbuf (PopplerPage *page,
+				int src_x, int src_y,
+				int src_width, int src_height,
+				double scale,
+				int rotation,
+				GBool printing,
+				GdkPixbuf *pixbuf)
+{
+  OutputDevData data;
+  
+  poppler_page_prepare_output_dev (page, scale, rotation, FALSE, &data);
+
+  page->page->displaySlice(page->document->output_dev,
+			   72.0 * scale, 72.0 * scale,
+			   rotation,
+			   gFalse, /* useMediaBox */
+			   gTrue, /* Crop */
+			   src_x, src_y,
+			   src_width, src_height,
+			   printing,
+			   page->document->doc->getCatalog (),
+			   NULL, NULL,
+			   printing ? poppler_print_annot_cb : NULL, NULL);
+  
+  poppler_page_copy_to_pixbuf (page, pixbuf, &data);
+}
+
+/**
+ * poppler_page_render_to_pixbuf:
+ * @page: the page to render from
+ * @src_x: x coordinate of upper left corner  
+ * @src_y: y coordinate of upper left corner  
+ * @src_width: width of rectangle to render  
+ * @src_height: height of rectangle to render
+ * @scale: scale specified as pixels per point
+ * @rotation: rotate the document by the specified degree
+ * @pixbuf: pixbuf to render into
+ *
+ * First scale the document to match the specified pixels per point,
+ * then render the rectangle given by the upper left corner at
+ * (src_x, src_y) and src_width and src_height.
+ * This function is for rendering a page that will be displayed.
+ * If you want to render a page that will be printed use
+ * poppler_page_render_to_pixbuf_for_printing() instead
+ **/
+void
+poppler_page_render_to_pixbuf (PopplerPage *page,
+			       int src_x, int src_y,
+			       int src_width, int src_height,
+			       double scale,
+			       int rotation,
+			       GdkPixbuf *pixbuf)
+{
+  g_return_if_fail (POPPLER_IS_PAGE (page));
+  g_return_if_fail (scale > 0.0);
+  g_return_if_fail (pixbuf != NULL);
+
+  _poppler_page_render_to_pixbuf (page, src_x, src_y,
+				  src_width, src_height,
+				  scale, rotation,
+				  gFalse,
+				  pixbuf);
+}
+
+/**
+ * poppler_page_render_to_pixbuf_for_printing:
+ * @page: the page to render from
+ * @src_x: x coordinate of upper left corner  
+ * @src_y: y coordinate of upper left corner  
+ * @src_width: width of rectangle to render  
+ * @src_height: height of rectangle to render
+ * @scale: scale specified as pixels per point
+ * @rotation: rotate the document by the specified degree
+ * @pixbuf: pixbuf to render into
+ *
+ * First scale the document to match the specified pixels per point,
+ * then render the rectangle given by the upper left corner at
+ * (src_x, src_y) and src_width and src_height.
+ * This function is for rendering a page that will be printed.
+ **/
+void
+poppler_page_render_to_pixbuf_for_printing (PopplerPage *page,
+					    int src_x, int src_y,
+					    int src_width, int src_height,
+					    double scale,
+					    int rotation,
+					    GdkPixbuf *pixbuf)
+{
+  g_return_if_fail (POPPLER_IS_PAGE (page));
+  g_return_if_fail (scale > 0.0);
+  g_return_if_fail (pixbuf != NULL);
+
+  _poppler_page_render_to_pixbuf (page, src_x, src_y,
+				  src_width, src_height,
+				  scale, rotation,
+				  gTrue,
+				  pixbuf);
+}
+
+/**
+ * poppler_page_get_thumbnail_pixbuf:
+ * @page: the #PopperPage to get the thumbnail for
+ * 
+ * Get the embedded thumbnail for the specified page.  If the document
+ * doesn't have an embedded thumbnail for the page, this function
+ * returns %NULL.
+ * 
+ * Return value: the tumbnail as a #GdkPixbuf or %NULL if the document
+ * doesn't have a thumbnail for this page.
+ **/
+GdkPixbuf *
+poppler_page_get_thumbnail_pixbuf (PopplerPage *page)
+{
+  unsigned char *data;
+  int width, height, rowstride;
+
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
+
+  if (!page->page->loadThumb (&data, &width, &height, &rowstride))
+    return NULL;
+
+  return gdk_pixbuf_new_from_data (data, GDK_COLORSPACE_RGB,
+				   FALSE, 8, width, height, rowstride,
+				   (GdkPixbufDestroyNotify)gfree, NULL);
+}
 
 /**
  * poppler_page_render_selection_to_pixbuf:
@@ -817,39 +942,7 @@ poppler_page_render_selection_to_pixbuf (PopplerPage           *page,
    * time. */
 }
 
-
-static void
-destroy_thumb_data (guchar *pixels, gpointer data)
-{
-  gfree (pixels);
-}
-
-/**
- * poppler_page_get_thumbnail:
- * @page: the #PopperPage to get the thumbnail for
- * 
- * Get the embedded thumbnail for the specified page.  If the document
- * doesn't have an embedded thumbnail for the page, this function
- * returns %NULL.
- * 
- * Return value: the tumbnail as a #GdkPixbuf or %NULL if the document
- * doesn't have a thumbnail for this page.
- **/
-GdkPixbuf *
-poppler_page_get_thumbnail (PopplerPage *page)
-{
-  unsigned char *data;
-  int width, height, rowstride;
-
-  g_return_val_if_fail (POPPLER_IS_PAGE (page), FALSE);
-
-  if (!page->page->loadThumb (&data, &width, &height, &rowstride))
-    return NULL;
-
-  return gdk_pixbuf_new_from_data (data, GDK_COLORSPACE_RGB,
-				   FALSE, 8, width, height, rowstride,
-				   destroy_thumb_data, NULL);
-}
+#endif /* POPPLER_WITH_GDK */
 
 /**
  * poppler_page_get_thumbnail_size:
@@ -897,6 +990,87 @@ poppler_page_get_thumbnail_size (PopplerPage *page,
   return retval;
 }
 
+/**
+ * poppler_page_get_selection_region:
+ * @page: a #PopplerPage
+ * @scale: scale specified as pixels per point
+ * @style: a #PopplerSelectionStyle
+ * @selection: start and end point of selection as a rectangle
+ * 
+ * Returns a region containing the area that would be rendered by
+ * poppler_page_render_selection() or 
+ * poppler_page_render_selection_to_pixbuf() as a #GList of
+ * #PopplerRectangle. The returned list must be freed with
+ * poppler_page_selection_region_free().
+ * 
+ * Return value: a #GList of #PopplerRectangle
+ **/
+GList *
+poppler_page_get_selection_region (PopplerPage           *page,
+				   gdouble                scale,
+				   PopplerSelectionStyle  style,
+				   PopplerRectangle      *selection)
+{
+  TextOutputDev *text_dev;
+  PDFRectangle poppler_selection;
+  SelectionStyle selection_style = selectionStyleGlyph;
+  GooList *list;
+  GList *region = NULL;
+  int i;
+
+  poppler_selection.x1 = selection->x1;
+  poppler_selection.y1 = selection->y1;
+  poppler_selection.x2 = selection->x2;
+  poppler_selection.y2 = selection->y2;
+
+  switch (style)
+    {
+      case POPPLER_SELECTION_GLYPH:
+        selection_style = selectionStyleGlyph;
+	break;
+      case POPPLER_SELECTION_WORD:
+        selection_style = selectionStyleWord;
+	break;
+      case POPPLER_SELECTION_LINE:
+        selection_style = selectionStyleLine;
+	break;
+    }
+	      
+  text_dev = poppler_page_get_text_output_dev (page);
+  list = text_dev->getSelectionRegion(&poppler_selection, 
+				      selection_style, scale);
+
+  for (i = 0; i < list->getLength(); i++) {
+    PDFRectangle *selection_rect = (PDFRectangle *) list->get(i);
+    PopplerRectangle *rect;
+
+    rect = poppler_rectangle_new ();
+    
+    rect->x1 = selection_rect->x1;
+    rect->y1 = selection_rect->y1;
+    rect->x2 = selection_rect->x2;
+    rect->y2 = selection_rect->y2;
+    
+    region = g_list_prepend (region, rect);
+    
+    delete selection_rect;
+  }
+
+  delete list;
+
+  return g_list_reverse (region);
+}
+
+void
+poppler_page_selection_region_free (GList *region)
+{
+  if (!region)
+    return;
+
+  g_list_foreach (region, (GFunc)poppler_rectangle_free, NULL);
+  g_list_free (region);
+}
+
 /**
  * poppler_page_get_text:
  * @page: a #PopplerPage
@@ -914,9 +1088,8 @@ poppler_page_get_text (PopplerPage          *page,
 		       PopplerRectangle     *selection)
 {
   TextOutputDev *text_dev;
-  PDFDoc *doc;
   GooString *sel_text;
-  double height, y1, y2;
+  double height;
   char *result;
   SelectionStyle selection_style = selectionStyleGlyph;
   PDFRectangle pdf_selection;
@@ -1014,46 +1187,32 @@ poppler_page_find_text (PopplerPage *page,
 #if defined (HAVE_CAIRO)
 
 static CairoImageOutputDev *
-poppler_page_get_image_output_dev (PopplerPage *page)
+poppler_page_get_image_output_dev (PopplerPage *page,
+				   GBool (*imgDrawDeviceCbk)(int img_id, void *data),
+				   void *imgDrawCbkData)
 {
-  if (page->image_dev == NULL) {
-    page->image_dev = new CairoImageOutputDev ();
+  CairoImageOutputDev *image_dev;
+  Gfx *gfx;
+  
+  image_dev = new CairoImageOutputDev ();
 
-    if (page->gfx)
-      delete page->gfx;
-    page->gfx = page->page->createGfx(page->image_dev,
-				      72.0, 72.0, 0,
-				      gFalse, /* useMediaBox */
-				      gTrue, /* Crop */
-				      -1, -1, -1, -1,
-				      gFalse, /* printing */
-				      page->document->doc->getCatalog (),
-				      NULL, NULL, NULL, NULL);
-    page->page->display(page->gfx);
+  if (imgDrawDeviceCbk) {
+    image_dev->setImageDrawDecideCbk (imgDrawDeviceCbk,
+				      imgDrawCbkData);
   }
 
-  return page->image_dev;
-}
+  gfx = page->page->createGfx(image_dev,
+			      72.0, 72.0, 0,
+			      gFalse, /* useMediaBox */
+			      gTrue, /* Crop */
+			      -1, -1, -1, -1,
+			      gFalse, /* printing */
+			      page->document->doc->getCatalog (),
+			      NULL, NULL, NULL, NULL);
+  page->page->display(gfx);
+  delete gfx;
 
-static GdkPixbuf *
-poppler_page_image_pixbuf_create (PopplerPage *page,
-				  CairoImage  *image)
-{
-  GdkPixbuf *pixbuf;
-  cairo_surface_t *surface;
-
-  surface = image->getImage ();
-  
-  pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
-			   FALSE, 8,
-			   cairo_image_surface_get_width (surface),
-			   cairo_image_surface_get_height (surface));
-  
-  copy_cairo_surface_to_pixbuf (surface,
-				cairo_image_surface_get_data (surface),
-				pixbuf);
-
-  return pixbuf;
+  return image_dev;
 }
 
 /**
@@ -1061,7 +1220,7 @@ poppler_page_image_pixbuf_create (PopplerPage *page,
  * @page: A #PopplerPage
  *
  * Returns a list of #PopplerImageMapping items that map from a
- * location on @page to a #GdkPixbuf.  This list must be freed
+ * location on @page to an image of the page. This list must be freed
  * with poppler_page_free_image_mapping() when done.
  *
  * Return value: A #GList of #PopplerImageMapping
@@ -1075,7 +1234,7 @@ poppler_page_get_image_mapping (PopplerPage *page)
   
   g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
 
-  out = poppler_page_get_image_output_dev (page);
+  out = poppler_page_get_image_output_dev (page, NULL, NULL);
 
   for (i = 0; i < out->getNumImages (); i++) {
     PopplerImageMapping *mapping;
@@ -1088,7 +1247,7 @@ poppler_page_get_image_mapping (PopplerPage *page)
 
     image->getRect (&(mapping->area.x1), &(mapping->area.y1),
 		    &(mapping->area.x2), &(mapping->area.y2));
-    mapping->image = poppler_page_image_pixbuf_create (page, image);
+    mapping->image_id = i;
     
     mapping->area.x1 -= page->page->getCropBox()->x1;
     mapping->area.x2 -= page->page->getCropBox()->x1;
@@ -1098,14 +1257,56 @@ poppler_page_get_image_mapping (PopplerPage *page)
     map_list = g_list_prepend (map_list, mapping);
   }
 
+  delete out;
+
   return map_list;	
 }
 
-static void
-poppler_images_mapping_free (PopplerImageMapping *mapping)
+static GBool
+image_draw_decide_cb (int image_id, void *data)
 {
-  g_object_unref (mapping->image);
-  g_free (mapping);
+  return (image_id == GPOINTER_TO_INT (data));
+}
+
+/**
+ * poppler_page_get_image:
+ * @page: A #PopplerPage
+ * @image_id: The image identificator
+ *
+ * Returns a cairo surface for the image of the @page
+ *
+ * Return value: A cairo surface for the image
+ **/
+cairo_surface_t *
+poppler_page_get_image (PopplerPage *page,
+			gint         image_id)
+{
+  CairoImageOutputDev *out;
+  cairo_surface_t *image;
+  
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+
+  out = poppler_page_get_image_output_dev (page,
+					   image_draw_decide_cb,
+					   GINT_TO_POINTER (image_id));
+
+  if (image_id >= out->getNumImages ()) {
+    delete out;
+    
+    return NULL;
+  }
+
+  image = out->getImage (image_id)->getImage ();
+  if (!image) {
+    delete out;
+
+    return NULL;
+  }
+
+  cairo_surface_reference (image);
+  delete out;
+  
+  return image;
 }
 
 /**
@@ -1121,11 +1322,11 @@ poppler_page_free_image_mapping (GList *list)
   if (list == NULL)
     return;
 
-  g_list_foreach (list, (GFunc) (poppler_images_mapping_free), NULL);
+  g_list_foreach (list, (GFunc)g_free, NULL);
   g_list_free (list);
 }
 
-#else
+#else /* HAVE_CAIRO */
 
 GList *
 poppler_page_get_image_mapping (PopplerPage *page)
@@ -1390,6 +1591,124 @@ poppler_page_free_form_field_mapping (GList *list)
   g_list_free (list);
 }
 
+/**
+ * poppler_page_get_annot_mapping:
+ * @page: A #PopplerPage
+ *
+ * Returns a list of #PopplerAnnotMapping items that map from a location on
+ * @page to a #PopplerAnnot.  This list must be freed with
+ * poppler_page_free_annot_mapping() when done.
+ *
+ * Return value: A #GList of #PopplerAnnotMapping
+ **/
+GList *
+poppler_page_get_annot_mapping (PopplerPage *page)
+{
+  GList *map_list = NULL;
+  double width, height;
+  gint i;
+
+  g_return_val_if_fail (POPPLER_IS_PAGE (page), NULL);
+
+  if (!page->annots)
+    page->annots = page->page->getAnnots (page->document->doc->getCatalog ());
+  
+  if (!page->annots)
+    return NULL;
+
+  poppler_page_get_size (page, &width, &height);
+
+  for (i = 0; i < page->annots->getNumAnnots (); i++) {
+    PopplerAnnotMapping *mapping;
+    PopplerRectangle rect;
+    Annot *annot;
+    PDFRectangle *annot_rect;
+    gint rotation = 0;
+
+    annot = page->annots->getAnnot (i);
+
+    /* Create the mapping */
+    mapping = poppler_annot_mapping_new ();
+
+    switch (annot->getType ())
+      {
+      case Annot::typeText:
+        mapping->annot = _poppler_annot_text_new (annot);
+	break;
+      case Annot::typeFreeText:
+        mapping->annot = _poppler_annot_free_text_new (annot);
+	break;
+      default:
+        mapping->annot = _poppler_annot_new (annot);
+	break;
+      }
+
+    annot_rect = annot->getRect ();
+    rect.x1 = annot_rect->x1;
+    rect.y1 = annot_rect->y1;
+    rect.x2 = annot_rect->x2;
+    rect.y2 = annot_rect->y2;
+
+    if (! (annot->getFlags () & Annot::flagNoRotate))
+      rotation = page->page->getRotate ();
+
+    switch (rotation)
+      {
+      case 90:
+        mapping->area.x1 = rect.y1;
+        mapping->area.y1 = height - rect.x2;
+        mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
+        mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
+        break;
+      case 180:
+        mapping->area.x1 = width - rect.x2;
+        mapping->area.y1 = height - rect.y2;
+        mapping->area.x2 = mapping->area.x1 + (rect.x2 - rect.x1);
+        mapping->area.y2 = mapping->area.y1 + (rect.y2 - rect.y1);
+        break;
+      case 270:
+        mapping->area.x1 = width - rect.y2;
+        mapping->area.y1 = rect.x1;
+        mapping->area.x2 = mapping->area.x1 + (rect.y2 - rect.y1);
+        mapping->area.y2 = mapping->area.y1 + (rect.x2 - rect.x1);
+        break;
+      default:
+        mapping->area.x1 = rect.x1;
+        mapping->area.y1 = rect.y1;
+        mapping->area.x2 = rect.x2;
+        mapping->area.y2 = rect.y2;
+      }
+
+    mapping->area.x1 -= page->page->getCropBox()->x1;
+    mapping->area.x2 -= page->page->getCropBox()->x1;
+    mapping->area.y1 -= page->page->getCropBox()->y1;
+    mapping->area.y2 -= page->page->getCropBox()->y1;
+
+    map_list = g_list_prepend (map_list, mapping);
+  }
+
+  return g_list_reverse (map_list);
+}
+
+/**
+ * poppler_page_free_annot_mapping:
+ * @list: A list of #PopplerAnnotMapping<!-- -->s
+ *
+ * Frees a list of #PopplerAnnotMapping<!-- -->s allocated by
+ * poppler_page_get_annot_mapping().  It also frees the #PopplerAnnot<!-- -->s
+ * that each mapping contains, so if you want to keep them around, you need to
+ * copy them with poppler_annot_copy().
+ **/
+void
+poppler_page_free_annot_mapping (GList *list)
+{
+  if (!list)
+    return;
+
+  g_list_foreach (list, (GFunc)poppler_annot_mapping_free, NULL);
+  g_list_free (list);
+}
+
 /* PopplerRectangle type */
 
 GType
@@ -1397,7 +1716,7 @@ poppler_rectangle_get_type (void)
 {
   static GType our_type = 0;
 
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static ("PopplerRectangle",
 					     (GBoxedCopyFunc) poppler_rectangle_copy,
 					     (GBoxedFreeFunc) poppler_rectangle_free);
@@ -1418,7 +1737,7 @@ poppler_rectangle_copy (PopplerRectangle *rectangle)
 
   g_return_val_if_fail (rectangle != NULL, NULL);
   
-  new_rectangle = g_new0 (PopplerRectangle, 1);
+  new_rectangle = g_new (PopplerRectangle, 1);
   *new_rectangle = *rectangle;
 
   return new_rectangle;
@@ -1430,13 +1749,50 @@ poppler_rectangle_free (PopplerRectangle *rectangle)
   g_free (rectangle);
 }
 
+/* PopplerColor type */
+GType
+poppler_color_get_type (void)
+{
+  static GType our_type = 0;
+
+  if (G_UNLIKELY (our_type == 0))
+    our_type = g_boxed_type_register_static ("PopplerColor",
+					     (GBoxedCopyFunc) poppler_color_copy,
+					     (GBoxedFreeFunc) poppler_color_free);
+
+  return our_type;
+}
+
+PopplerColor *
+poppler_color_new (void)
+{
+  return (PopplerColor *) g_new0 (PopplerColor, 1);
+}
+
+PopplerColor *
+poppler_color_copy (PopplerColor *color)
+{
+  PopplerColor *new_color;
+
+  new_color = g_new (PopplerColor, 1);
+  *new_color = *color;
+
+  return new_color;
+}
+
+void
+poppler_color_free (PopplerColor *color)
+{
+  g_free (color);
+}
+
 /* PopplerLinkMapping type */
 GType
 poppler_link_mapping_get_type (void)
 {
   static GType our_type = 0;
 
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static ("PopplerLinkMapping",
 					     (GBoxedCopyFunc) poppler_link_mapping_copy,
 					     (GBoxedFreeFunc) poppler_link_mapping_free);
@@ -1479,7 +1835,7 @@ poppler_image_mapping_get_type (void)
 {
   static GType our_type = 0;
 
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static ("PopplerImageMapping",
 					     (GBoxedCopyFunc) poppler_image_mapping_copy,
 					     (GBoxedFreeFunc) poppler_image_mapping_free);
@@ -1501,8 +1857,6 @@ poppler_image_mapping_copy (PopplerImageMapping *mapping)
   new_mapping = poppler_image_mapping_new ();
 
   *new_mapping = *mapping;
-  if (new_mapping->image)
-    new_mapping->image = gdk_pixbuf_copy (new_mapping->image);
 
   return new_mapping;
 }
@@ -1510,9 +1864,6 @@ poppler_image_mapping_copy (PopplerImageMapping *mapping)
 void
 poppler_image_mapping_free (PopplerImageMapping *mapping)
 {
-  if (mapping->image)
-    g_object_unref (mapping->image);
-
   g_free (mapping);
 }
 
@@ -1521,7 +1872,7 @@ GType
 poppler_page_transition_get_type (void)
 {
   static GType our_type = 0;
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static("PopplerPageTransition",
 					    (GBoxedCopyFunc) poppler_page_transition_copy,
 					    (GBoxedFreeFunc) poppler_page_transition_free);
@@ -1556,7 +1907,7 @@ GType
 poppler_form_field_mapping_get_type (void)
 {
   static GType our_type = 0;
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static("PopplerFormFieldMapping",
 					    (GBoxedCopyFunc) poppler_form_field_mapping_copy,
 					    (GBoxedFreeFunc) poppler_form_field_mapping_free);
@@ -1595,6 +1946,52 @@ poppler_form_field_mapping_free (PopplerFormFieldMapping *mapping)
   g_free (mapping);
 }
 
+/* PopplerAnnot Mapping Type */
+GType
+poppler_annot_mapping_get_type (void)
+{
+  static GType our_type = 0;
+
+  if (our_type == 0)
+    our_type = g_boxed_type_register_static ("PopplerAnnotMapping",
+					     (GBoxedCopyFunc) poppler_annot_mapping_copy,
+					     (GBoxedFreeFunc) poppler_annot_mapping_free);
+
+  return our_type;
+}
+
+PopplerAnnotMapping *
+poppler_annot_mapping_new (void)
+{
+  return (PopplerAnnotMapping *) g_new0 (PopplerAnnotMapping, 1);
+}
+
+PopplerAnnotMapping *
+poppler_annot_mapping_copy (PopplerAnnotMapping *mapping)
+{
+  PopplerAnnotMapping *new_mapping;
+
+  new_mapping = poppler_annot_mapping_new ();
+
+  *new_mapping = *mapping;
+  if (mapping->annot)
+    new_mapping->annot = (PopplerAnnot *) g_object_ref (mapping->annot);
+
+  return new_mapping;
+}
+
+void
+poppler_annot_mapping_free (PopplerAnnotMapping *mapping)
+{
+  if (!mapping)
+    return;
+
+  if (mapping->annot)
+    g_object_unref (mapping->annot);
+
+  g_free (mapping);
+}
+
 void 
 poppler_page_get_crop_box (PopplerPage *page, PopplerRectangle *rect)
 {
