@@ -412,6 +412,7 @@ GBool XRef::readXRefTable(Parser *parser, Guint *pos) {
 	entries[i].offset = 0xffffffff;
 	entries[i].type = xrefEntryFree;
 	entries[i].obj.initNull ();
+	entries[i].updated = false;
       }
       size = newSize;
     }
@@ -524,6 +525,7 @@ GBool XRef::readXRefStream(Stream *xrefStr, Guint *pos) {
       entries[i].offset = 0xffffffff;
       entries[i].type = xrefEntryFree;
       entries[i].obj.initNull ();
+      entries[i].updated = false;
     }
     size = newSize;
   }
@@ -619,6 +621,7 @@ GBool XRef::readXRefStreamSection(Stream *xrefStr, int *w, int first, int n) {
       entries[i].offset = 0xffffffff;
       entries[i].type = xrefEntryFree;
       entries[i].obj.initNull ();
+      entries[i].updated = false;
     }
     size = newSize;
   }
@@ -754,16 +757,17 @@ GBool XRef::constructXRef() {
 		    error(-1, "Bad object number");
 		    return gFalse;
 		  }
-                  if (newSize*(int)sizeof(XRefEntry)/sizeof(XRefEntry) != newSize) {
-                    error(-1, "Invalid 'obj' parameters.");
-                    return gFalse;
-                  }
+		  if (newSize*(int)sizeof(XRefEntry)/sizeof(XRefEntry) != newSize) {
+		    error(-1, "Invalid 'obj' parameters.");
+		    return gFalse;
+		  }
 		  entries = (XRefEntry *)
 		      greallocn(entries, newSize, sizeof(XRefEntry));
 		  for (i = size; i < newSize; ++i) {
 		    entries[i].offset = 0xffffffff;
 		    entries[i].type = xrefEntryFree;
 		    entries[i].obj.initNull ();
+		    entries[i].updated = false;
 		  }
 		  size = newSize;
 		}
@@ -1006,6 +1010,7 @@ void XRef::add(int num, int gen, Guint offs, GBool used) {
   e->gen = gen;
   e->num = num;
   e->obj.initNull ();
+  e->updated = false;
   if (used) {
     e->type = xrefEntryUncompressed;
     e->offset = offs;
@@ -1021,15 +1026,56 @@ void XRef::setModifiedObject (Object* o, Ref r) {
     return;
   }
   o->copy(&entries[r.num].obj);
+  entries[r.num].updated = true;
 }
 
+Ref XRef::addIndirectObject (Object* o) {
+  //Find the next free entry
+  // lastEntry is the last free entry we encountered,
+  // entry 0 is always free
+  int lastEntry = 0;
+  int newEntry = entries[0].offset;
+
+  do {
+    lastEntry = newEntry;
+    newEntry = entries[newEntry].offset;
+    //we are looking for a free entry that we can reuse
+    //(=> gen number != 65535)
+  } while ((newEntry != 0) && (entries[newEntry].gen == 65535));
+
+  //the linked list of free entry is empty => create a new one
+  if (newEntry == 0) {
+    newEntry = size;
+    size++;
+    entries = (XRefEntry *)greallocn(entries, size, sizeof(XRefEntry));
+    entries[newEntry].gen = 0;
+    entries[newEntry].num = newEntry;
+  } else { //reuse a free entry
+    //'remove' the entry we are using from the free entry linked list
+    entries[lastEntry].offset = entries[newEntry].offset;
+    entries[newEntry].num = newEntry;
+    //we don't touch gen number, because it should have been 
+    //incremented when the object was deleted
+  }
+
+  entries[newEntry].type = xrefEntryUncompressed;
+  o->copy(&entries[newEntry].obj);
+  entries[newEntry].updated = true;
+
+  Ref r;
+  r.num = entries[newEntry].num;
+  r.gen = entries[newEntry].gen;
+  return r;
+}
+
+
 //used to sort the entries
 int compare (const void* a, const void* b)
 {
   return (((XRefEntry*)a)->num - ((XRefEntry*)b)->num);
 }
 
-void XRef::writeToFile(FILE* file) {
+void XRef::writeToFile(OutStream* outStr) {
   qsort(entries, size, sizeof(XRefEntry), compare);
   //create free entries linked-list
   if (entries[0].gen != 65535) {
@@ -1044,16 +1090,16 @@ void XRef::writeToFile(FILE* file) {
   }
   //write the new xref
   int j;
-  fprintf(file,"xref\r\n");
+  outStr->printf("xref\r\n");
   for (int i=0; i<size; i++) {
-    for(j=i; j<size; j++) { //look for consecutive entry
+    for(j=i; j<size; j++) { //look for consecutive entries
       if (j!=i && entries[j].num != entries[j-1].num+1) 
               break;
     }
-    fprintf(file,"%i %i\r\n", entries[i].num, j-i);
+    outStr->printf("%i %i\r\n", entries[i].num, j-i);
     for (int k=i; k<j; k++) {
       if(entries[k].gen > 65535) entries[k].gen = 65535; //cap generation number to 65535 (required by PDFReference)
-      fprintf(file,"%010i %05i %c\r\n", entries[k].offset, entries[k].gen, (entries[k].type==xrefEntryFree)?'f':'n');
+      outStr->printf("%010i %05i %c\r\n", entries[k].offset, entries[k].gen, (entries[k].type==xrefEntryFree)?'f':'n');
     }
     i = j-1;
   }
